# List of all the observatory package names
ALL_OBSERVATORIES = ["hst", "jwst", "roman", "tobs"]

# keywords used to identify instrument from headers
INSTRUMENT_KEYWORDS = ["INSTRUME", "META.INSTRUMENT.NAME",
                       "META_INSTRUMENT_NAME", "INSTRUMENT",
                       "ROMAN.META.INSTRUMENT.NAME",
                       "META.INSTRUMENT.TYPE", "META_INSTRUMENT_TYPE"]
