# This environment is used for running unit tests as stand alone modules
# rather than under the control of runtests.

export CRDS_SERVER_URL=https://hst-crds.stsci.edu

export CRDS_TEST_ROOT="$HOME"

export CRDS_PATH="${CRDS_TEST_ROOT}/crds-cache-default-test"

export CRDS_CACHE_DEFAULT="$CRDS_PATH"

export CRDS_TESTING_CACHE="${CRDS_TEST_ROOT}/crds-cache-test"

export CRDS_GROUP_CACHE="$HOME/crds_cache_ops"

