unset CRDS_SERVER_URL
unset CRDS_PATH
unset CRDS_S3_ENABLED
unset CRDS_S3_RETURN_URI
unset CRDS_MAPPING_URI
unset CRDS_REFERENCE_URI
unset CRDS_CONFIG_URI

unset CRDS_USE_PICKLED_CONTEXTS
unset CRDS_PICKLE_URI

unset CRDS_DOWNLOAD_MODE
unset CRDS_DOWNLOAD_PLUGIN
